/* dosfd_img.c version 2.0.0 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "dosfd_img.h"


static int get_int2(const unsigned char *data);
static int get_int4(const unsigned char *data);
static int make_entry_name(char *entry_name, const char *file_name);
static int get_sector_from_cluster(const struct dosfd_image_t *image, int cluster);
static int get_cluster_from_sector(const struct dosfd_image_t *image, int sector, int *mod);
static void get_CHR_from_clustor(const struct dosfd_image_t *image, int cluster, int *c, int *h, int *r);
static int get_cluster_from_CHR(const struct dosfd_image_t *image, int c, int h, int r, int *mod);
static int get_image_offset(const struct dosfd_image_t *image, int sector);
static int get_fat(struct dosfd_image_t *image);
static int get_next_cluster(const struct dosfd_image_t *image, int cluster);
static const unsigned char *find_file_entry(const struct dosfd_image_t *image, int cluster, int num_entries, const char *entry_name, int ignore_case);


static int get_int2(const unsigned char *data)
{
	int val;

	val = (int)*data++;
	val |= (int)(*data++) << 8;

	return val;
}

static int get_int4(const unsigned char *data)
{
	int val;

	val = (int)*data++;
	val |= (int)(*data++) << 8;
	val |= (int)(*data++) << 16;
	val |= (int)(*data++) << 24;

	return val;
}

static int make_entry_name(char *entry_name, const char *file_name)
{
	const char *src = file_name;
	int i;

	for (i = 0; i < 8 && *src && *src != '.'; i++, src++) {
		entry_name[i] = *src;
	}
	for ( ; i < 8; i++) {
		entry_name[i] = ' ';
	}

	if (*src == '.') {
		src++;
		for ( ; i < 11 && *src; i++, src++) {
			entry_name[i] = *src;
		}
	}
	for ( ; i < 11; i++) {
		entry_name[i] = ' ';
	}
	entry_name[i] = '\0';

	if (entry_name[0] == (char)0xE5) {
		entry_name[0] = 0x05;
	}

	if (*src) {
		return -1;
	}

	return 0;
}

static int get_sector_from_cluster(const struct dosfd_image_t *image, int cluster)
{
	if (cluster < 2 || image->fat_size <= cluster) {
		return -1;
	}

	return image->cluster_2_sector + (cluster - 2) * image->sectors_per_cluster;
}

static int get_cluster_from_sector(const struct dosfd_image_t *image, int sector, int *mod)
{	
	int cluster;

	sector -= image->cluster_2_sector;
	cluster = sector / image->sectors_per_cluster + 2;

	if (mod != NULL) {
		*mod = sector % image->sectors_per_cluster;
	}

	return cluster;
}

static void get_CHR_from_clustor(const struct dosfd_image_t *image, int cluster, int *c, int *h, int *r)
{
	int sector = get_sector_from_cluster(image, cluster);

	*c = sector / (image->sectors_per_track * image->num_heads);
	sector -= *c * (image->sectors_per_track * image->num_heads);
	*h = sector / image->sectors_per_track;
	sector -= *h * image->sectors_per_track;
	*r = sector + 1;
}

static int get_cluster_from_CHR(const struct dosfd_image_t *image, int c, int h, int r, int *mod)
{
	int sector = (c * image->num_heads + h) * image->sectors_per_track + r - 1;

	return get_cluster_from_sector(image, sector, mod);
}

static int get_image_offset(const struct dosfd_image_t *image, int sector)
{
	return image->img_offset + sector * image->sector_size;
}

static int get_fat(struct dosfd_image_t *image)
{
	int result = -1;
	unsigned char *fat = NULL;
	unsigned char *ptr;
	int i;

	if (image == NULL || image->disk_img == NULL) {
		return -1;
	}

	image->fat_size = (image->num_sectors - image->cluster_2_sector) / image->sectors_per_cluster;
	if ((image->sectors_per_fat * image->sector_size * 2 / 3) < image->fat_size) {
		printf("can't calculate FAT size\n");
		goto ERROR_EXIT;
	}

	if ((fat = (unsigned char *)malloc(image->sectors_per_fat * image->sector_size)) == NULL) {
		goto ERROR_EXIT;
	}

	for (i = 0, ptr = fat; i < image->sectors_per_fat; i++, ptr += image->sector_size) {
		memcpy(ptr,
			   image->disk_img + get_image_offset(image, image->fat_sector + i),
			   image->sector_size);
	}

	if ((image->fat = (int *)malloc(sizeof(int) * image->fat_size)) == NULL) {
		printf("memory exhausted!");
		goto ERROR_EXIT;
	}

	for (i = 0, ptr = fat; i < image->fat_size; i += 2) {
		int val;
		val = (int)*ptr++;
		val |= ((int)*ptr++) << 8;
		val |= ((int)*ptr++) << 16;
		image->fat[i    ] = val & 0xFFF;
		image->fat[i + 1] = val >> 12;
	}

	result = 0;

ERROR_EXIT:
	if (fat != NULL) {
		free(fat);
	}

	return result;
}

static int get_next_cluster(const struct dosfd_image_t *image, int cluster)
{
	if (cluster < 0 || image->fat_size <= cluster) {
		return -1;
	}

	return image->fat[cluster];
}

static const unsigned char *find_file_entry(const struct dosfd_image_t *image, int cluster, int num_entries, const char *entry_name, int ignore_case)
{
	int sector;

	if (2 <= cluster) {
		sector = get_sector_from_cluster(image, cluster);
	} else {
		/* root directory */
		sector = image->root_dir_sector;
	}

	while (0 < num_entries) {
		int num_sectors = image->sectors_per_cluster;

		for ( ; 0 < num_entries && 0 < num_sectors--; sector++) {
			const unsigned char *ptr = image->disk_img + get_image_offset(image, sector);
			int count = image->sector_size / DOSFD_FILE_ENTRY_SIZE;

			for ( ; 0 < count-- && 0 < num_entries--; ptr += DOSFD_FILE_ENTRY_SIZE) {
				if ((ignore_case == 0 && strncmp((const char *)ptr, entry_name, 11) == 0) ||
					(ignore_case != 0 && strncasecmp((const char *)ptr, entry_name, 11) == 0)) {
					return ptr;
				}
			}
		}

		if (2 <= cluster) {
			cluster = get_next_cluster(image, cluster);
			if (cluster < 2 || 0xFF0 <= cluster) {
				break;
			}
			sector = get_sector_from_cluster(image, cluster);
		}
	}

	return NULL;
}

void free_dosfd_image(struct dosfd_image_t *image, int free_all)
{
	if (image->disk_img != NULL) {
		free(image->disk_img);
	}

	if (image->fat != NULL) {
		free(image->fat);
	}

	if (free_all) {
		free(image);
	} else {
		memset(image, 0, sizeof(struct dosfd_image_t));
	}
}

int read_dosfd_image(struct dosfd_image_t *image, const char *image_name, int img_offset)
{
	int result = -1;
	FILE *fp = NULL;

	memset(image, 0, sizeof(*image));
	image->img_offset = img_offset;

	if ((fp = fopen(image_name, "rb")) == NULL) {
		printf("open error '%s'\n", image_name);
		goto ERROR_EXIT;
	}

	if (fseek(fp, 0, SEEK_END) < 0) {
		printf("seek error '%s' (SEEK_END)\n", image_name);
		goto ERROR_EXIT;
	}
	image->disk_size = ftell(fp);

	if (fseek(fp, 0, SEEK_SET) < 0) {
		printf("seek error '%s' (SEEK_SET)\n", image_name);
		goto ERROR_EXIT;
	}

	if ((image->disk_img = (unsigned char *)malloc(image->disk_size)) == NULL) {
		printf("memory exhausted!\n");
		goto ERROR_EXIT;
	}

	if (fread(image->disk_img, image->disk_size, 1, fp) != 1) {
		printf("read error '%s'\n", image_name);
		goto ERROR_EXIT;
	}

	fclose(fp);
	fp = NULL;

	if (image->disk_size == img_offset + DOSFD_2HD_SIZE) {
		image->num_tracks           = DOSFD_2HD_TRACKS;
		image->num_heads            = DOSFD_2HD_HEADS;
		image->sectors_per_track    = DOSFD_2HD_SECTORS_PER_TRACK;
		image->sector_size          = DOSFD_2HD_SECTOR_SIZE;
		image->num_sectors          = DOSFD_2HD_SECTORS;
		image->fat_sector           = DOSFD_2HD_FAT_SECTOR;
		image->sectors_per_fat      = DOSFD_2HD_SECTORS_PER_FAT;
		image->root_dir_sector      = DOSFD_2HD_ROOT_DIR_SECTOR;
		image->root_dir_entries     = DOSFD_2HD_ROOT_DIR_ENTRIES;
		image->sectors_per_root_dir = DOSFD_2HD_SECTORS_PER_ROOT_DIR;
		image->cluster_2_sector     = DOSFD_2HD_CLUSTER_2_SECTOR;
		image->sectors_per_cluster  = DOSFD_2HD_SECTORS_PER_CLUSTER;
	} else if (image->disk_size == img_offset + DOSFD_2DD9_SIZE) {
		image->num_tracks           = DOSFD_2DD9_TRACKS;
		image->num_heads            = DOSFD_2DD9_HEADS;
		image->sectors_per_track    = DOSFD_2DD9_SECTORS_PER_TRACK;
		image->sector_size          = DOSFD_2DD9_SECTOR_SIZE;
		image->num_sectors          = DOSFD_2DD9_SECTORS;
		image->fat_sector           = DOSFD_2DD9_FAT_SECTOR;
		image->sectors_per_fat      = DOSFD_2DD9_SECTORS_PER_FAT;
		image->root_dir_sector      = DOSFD_2DD9_ROOT_DIR_SECTOR;
		image->root_dir_entries     = DOSFD_2DD9_ROOT_DIR_ENTRIES;
		image->sectors_per_root_dir = DOSFD_2DD9_SECTORS_PER_ROOT_DIR;
		image->cluster_2_sector     = DOSFD_2DD9_CLUSTER_2_SECTOR;
		image->sectors_per_cluster  = DOSFD_2DD9_SECTORS_PER_CLUSTER;
	} else if (image->disk_size == img_offset + DOSFD_2DD8_SIZE) {
		image->num_tracks           = DOSFD_2DD8_TRACKS;
		image->num_heads            = DOSFD_2DD8_HEADS;
		image->sectors_per_track    = DOSFD_2DD8_SECTORS_PER_TRACK;
		image->sector_size          = DOSFD_2DD8_SECTOR_SIZE;
		image->num_sectors          = DOSFD_2DD8_SECTORS;
		image->fat_sector           = DOSFD_2DD8_FAT_SECTOR;
		image->sectors_per_fat      = DOSFD_2DD8_SECTORS_PER_FAT;
		image->root_dir_sector      = DOSFD_2DD8_ROOT_DIR_SECTOR;
		image->root_dir_entries     = DOSFD_2DD8_ROOT_DIR_ENTRIES;
		image->sectors_per_root_dir = DOSFD_2DD8_SECTORS_PER_ROOT_DIR;
		image->cluster_2_sector     = DOSFD_2DD8_CLUSTER_2_SECTOR;
		image->sectors_per_cluster  = DOSFD_2DD8_SECTORS_PER_CLUSTER;
	} else {
		printf("unsupported image '%s'\n", image_name);
		goto ERROR_EXIT;
	}

	if (get_fat(image) < 0) {
		printf("can't get fat '%s' offset=%08X\n", image_name, img_offset);
		goto ERROR_EXIT;
	}

	result = 0;

ERROR_EXIT:
	if (fp != NULL) {
		fclose(fp);
	}
	if (result < 0) {
		free_dosfd_image(image, 0);
	}

	return result;
}

int get_dosfd_image_file_size(const struct dosfd_image_t *image, const char *file_name, int ignore_case)
{
	char entry_name[12];
	const unsigned char *ptr;

	if (make_entry_name(entry_name, file_name) < 0) {
		printf("illegal filename '%s'\n", file_name);
		return -1;
	}

	if ((ptr = find_file_entry(image, -1, image->root_dir_entries, entry_name, ignore_case)) == NULL) {
		printf("can't find '%s'\n", file_name);
		return -1;
	}

	return get_int4(ptr + DOSFD_FILE_ENTRY_OFFSET_SIZE);
}

int read_dosfd_image_file(const struct dosfd_image_t *image, const char *file_name, unsigned char *data, int data_size, int ignore_case)
{
	char entry_name[12];
	const unsigned char *ptr;
	int cluster;
	int entry_size;

	if (make_entry_name(entry_name, file_name) < 0) {
		printf("illegal filename '%s'\n", file_name);
		return -1;
	}

	if ((ptr = find_file_entry(image, -1, image->root_dir_entries, entry_name, ignore_case)) == NULL) {
		printf("can't find '%s'\n", file_name);
		return -1;
	}

	cluster = get_int2(ptr + DOSFD_FILE_ENTRY_OFFSET_CLUSTER);
	entry_size = get_int4(ptr + DOSFD_FILE_ENTRY_OFFSET_SIZE);

	if (data_size < entry_size) {
		printf("file size error '%s' %d > %d\n", file_name, entry_size, data_size);
		return -1;
	} else if (entry_size < data_size) {
		printf("file size mismatch '%s' %d != %d ... ignored\n", file_name, entry_size, data_size);
	}

	{
		unsigned char *des = data;
		int size = entry_size;
		int sector_size = image->sector_size;

		while (0 < size) {
			int num_sectors;
			int sector;

			if (cluster < 2 || 0xFF0 <= cluster) {
				printf("illegal cluster 0x%X\n", cluster);
				return -1;
			}

			sector = get_sector_from_cluster(image, cluster);
			if (sector < 0) {
				printf("can't get sector from cluster 0x%X\n", cluster);
				return -1;
			}

			num_sectors = image->sectors_per_cluster;
			while (0 < num_sectors-- && 0 < size) {
				int copy_size = sector_size < size ? sector_size : size;

				memcpy(des, image->disk_img + get_image_offset(image, sector), copy_size);

				size -= copy_size;
				des += copy_size;
				sector++;
			}

			cluster = get_next_cluster(image, cluster);
		}
	}

	return 0;
}

int write_dosfd_image(const struct dosfd_image_t *image, const char *image_name)
{
	int result = -1;
	FILE *fp = NULL;

	if ((fp = fopen(image_name, "wb")) == NULL) {
		printf("open error '%s'\n", image_name);
		goto ERROR_EXIT;
	}

	if (fwrite(image->disk_img, image->disk_size, 1, fp) != 1) {
		printf("write error '%s'\n", image_name);
		goto ERROR_EXIT;
	}

	fclose(fp);
	fp = NULL;

	result = 0;

ERROR_EXIT:
	if (fp != NULL) {
		fclose(fp);
	}

	return result;
}

int write_dosfd_image_file(struct dosfd_image_t *image, const char *file_name, const unsigned char *data, int data_size, int ignore_case)
{
	char entry_name[12];
	const unsigned char *ptr;
	int cluster;
	int entry_size;

	if (make_entry_name(entry_name, file_name) < 0) {
		printf("illegal filename '%s'\n", file_name);
		return -1;
	}

	if ((ptr = find_file_entry(image, -1, image->root_dir_entries, entry_name, ignore_case)) == NULL) {
		printf("can't find '%s'\n", file_name);
		return -1;
	}

	cluster = get_int2(ptr + DOSFD_FILE_ENTRY_OFFSET_CLUSTER);
	entry_size = get_int4(ptr + DOSFD_FILE_ENTRY_OFFSET_SIZE);

	if (entry_size < data_size) {
		printf("file size error '%s' %d < %d\n", file_name, entry_size, data_size);
		return -1;
	} else if (data_size < entry_size) {
		printf("file size mismatch '%s' %d != %d ... ignored\n", file_name, entry_size, data_size);
	}

	{
		const unsigned char *src = data;
		int size = data_size;
		int sector_size = image->sector_size;

		while (0 < size) {
			int num_sectors;
			int sector;

			if (cluster < 2 || 0xFF0 <= cluster) {
				printf("illegal cluster 0x%X\n", cluster);
				return -1;
			}

			sector = get_sector_from_cluster(image, cluster);
			if (sector < 0) {
				printf("can't get sector from cluster 0x%X\n", cluster);
				return -1;
			}

			num_sectors = image->sectors_per_cluster;
			while (0 < num_sectors-- && 0 < size) {
				int copy_size = sector_size < size ? sector_size : size;

				memcpy(image->disk_img + get_image_offset(image, sector), src, copy_size);

				size -= copy_size;
				src += copy_size;
				sector++;
			}

			cluster = get_next_cluster(image, cluster);
		}
	}

	return 0;
}

/* EOF */
