#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>

#ifdef U4_PORT_PC88
#include "d88image.h"
#endif	/* U4_PORT_PC88 */
#if defined(U4_PORT_PC98) || defined(U4_PORT_X68K)
#include "dosfd_img.h"
#endif	/* U4_PORT_PC98 || U4_PORT_X68K */


#define DNG_ID_NONE			(-1)
#define DNG_ID_DECEIT		(0)
#define DNG_ID_DESPISE		(1)
#define DNG_ID_DESTARD		(2)
#define DNG_ID_WRONG		(3)
#define DNG_ID_COVETOUS		(4)
#define DNG_ID_SHAME		(5)
#define DNG_ID_HYTHLOTH		(6)
#define DNG_ID_ABYSS		(7)
#define DNG_ID_MAX			(8)


#ifdef U4_PORT_PC88
struct sec_info_t {
	int cylinder;
	int head;
	int record;
};

static const struct sec_info_t sec_info_pc88[DNG_ID_MAX][2] = {
	{ { 0x01, 0, 0x01 }, { 0x01, 0, 0x02 } },			/* Deceit   */
	{ { 0x01, 0, 0x03 }, { 0x01, 0, 0x04 } },			/* Despise  */
	{ { 0x01, 0, 0x05 }, { 0x01, 0, 0x06 } },			/* Destard  */
	{ { 0x01, 0, 0x07 }, { 0x01, 0, 0x08 } },			/* Wrong    */
	{ { 0x01, 0, 0x09 }, { 0x01, 0, 0x0A } },			/* Covetous */
	{ { 0x01, 0, 0x0B }, { 0x01, 0, 0x0C } },			/* Shame    */
	{ { 0x01, 0, 0x0D }, { 0x01, 0, 0x0E } },			/* Hythloth */
	{ { 0x01, 0, 0x0F }, { 0x01, 0, 0x10 } }			/* Abyss    */
};

static const char *program_name = "u4pc88fixdunmap";
static const char *default_img_file = "u4-underworld.d88";
#endif	/* U4_PORT_PC88 */

#ifdef U4_PORT_PC98
static const char *program_name = "u4pc98fixdunmap";
static const char *default_img_file = "U4_Britania.fdi";
#endif	/* U4_PORT_PC98 */

#ifdef U4_PORT_X68K
static const char *program_name = "u4x68kfixdunmap";
static const char *default_img_file = "u4-britannia.xdf";
#endif	/* U4_PORT_X68K */


static void usage(void);
#ifdef U4_PORT_PC88
static int modify_map_data_d88(const char *img_file, const char *out_file, int image_index, const struct sec_info_t sec_info[DNG_ID_MAX][2], int opt_overwrite);
#endif	/* U4_PORT_PC88 */
#if defined(U4_PORT_PC98) || defined(U4_PORT_X68K)
static int modify_map_data_dosfd(const char *img_file, const char *out_file, int offset, int opt_overwrite);
#endif	/* U4_PORT_PC98 || U4_PORT_X68K */
static int modify_map_data(unsigned char map_data[8][8][8], int dng_id);
static void dump_map_data(unsigned char map_data[8][8]);
static char *create_new_name(const char *name);
static int check_out_file(const char *filename);
static int confirm(const char *msg);


int main(int argc, char **argv)
{
	int result = 1;
	int num_cmd_params = 1;

	const char *img_file = NULL;
	const char *out_file = NULL;
	char *new_name_buf = NULL;
#ifdef U4_PORT_PC88
	int image_index = 0;
#endif	/* U4_PORT_PC88 */
#ifdef U4_PORT_PC98
	int image_offset = FDI_OFFSET;
#endif	/* U4_PORT_PC98 */
#ifdef U4_PORT_X68K
	int image_offset = 0;
#endif	/* U4_PORT_X68K */
	int opt_overwrite = 0;

	while (*++argv != NULL) {
		if (**argv == '-') {
			if (strcmp(*argv, "-?") == 0 || strcmp(*argv, "--help") == 0) {
				usage();
				return 1;
#ifdef U4_PORT_PC88
			} else if (strcmp(*argv, "-i") == 0 || strcmp(*argv, "--index") == 0) {
				char *endptr;
				if (*++argv == NULL) {
					printf("no image index\n");
					return 1;
				}
				image_index = strtoul(*argv, &endptr, 0);
				if (endptr == NULL || *endptr != '\0') {
					printf("illegal image index '%s'\n", *argv);
					return 1;
				}
#endif	/* U4_PORT_PC88 */
			} else if (strcmp(*argv, "-o") == 0 || strcmp(*argv, "--output") == 0) {
				if (*++argv == NULL) {
					printf("no disk image\n");
					return 1;
				}
				out_file = *argv;
			} else if (strcmp(*argv, "--overwrite") == 0) {
				opt_overwrite = 1;
			} else {
				printf("unknown option '%s'\n", *argv);
				return 1;
			}
		} else {
			if (num_cmd_params <= 0) {
				printf("unneeded parameter '%s'\n", *argv);
				return 1;
			}
			img_file = *argv;
			num_cmd_params--;
		}
	}

	/* default image file */
	if (img_file == NULL) {
		img_file = default_img_file;
	}

	if (out_file == NULL) {
		if ((new_name_buf = create_new_name(img_file)) == NULL) {
			printf("ERROR: can't create output filename\n");
			goto ERROR_EXIT;
		}
		out_file = new_name_buf;
	}

	/* modify map data */
#ifdef U4_PORT_PC88
	if (modify_map_data_d88(img_file, out_file, image_index, sec_info_pc88, opt_overwrite) <= 0) {
		goto ERROR_EXIT;
	}
#endif	/* U4_PORT_PC88 */
#if defined(U4_PORT_PC98) || defined(U4_PORT_X68K)
	if (modify_map_data_dosfd(img_file, out_file, image_offset, opt_overwrite) <= 0) {
		goto ERROR_EXIT;
	}
#endif	/* U4_PORT_PC98 || U4_PORT_X68K */

	printf("completed!\n");

	result = 0;

ERROR_EXIT:
	if (new_name_buf != NULL) {
		free(new_name_buf);
	}

	return result;
}

static void usage(void)
{
	printf("usage: %s [options] [DISK-IMG]\n", program_name);
	printf("option:\n");
#ifdef U4_PORT_PC88
	printf("        -i|--index NUM            specify image index of disk image\n");
	printf("                                  default index is 0\n");
#endif	/* U4_PORT_PC88 */
	printf("        -o|--output MODIFIED-IMG  write modified disk to specified file\n");
	printf("           --overwrite            overwrite file without confirmation\n");
	printf("        -?|--help                 print usage\n");
	printf("\n");
#ifdef U4_PORT_PC88
	printf("supported disk image format is *.d88 only\n");
#endif	/* U4_PORT_PC88 */
#ifdef U4_PORT_PC98
	printf("supported disk image format is *.fdi only\n");
#endif	/* U4_PORT_PC98 */
#ifdef U4_PORT_X68K
	printf("supported disk image format is *.xdf only\n");
#endif	/* U4_PORT_X68K */
	printf("read disk image from '%s' if not specified\n", default_img_file);
	printf("write modified disk image to filename.NEW.ext if not specified\n");
}

#ifdef U4_PORT_PC88
static int modify_map_data_d88(const char *img_file, const char *out_file, int image_index, const struct sec_info_t sec_info[DNG_ID_MAX][2], int opt_overwrite)
{
	int result = -1;
	struct d88_image_t image;
	struct d88_disk_t *disk;
	unsigned char map_data[8][8][8];		/* map_data[level][y][x] */
	const int ids[2] = { DNG_ID_WRONG, DNG_ID_COVETOUS };
	int n;
	int modified = 0;

	memset(&image, 0, sizeof(image));

	if (read_d88_image(&image, img_file) < 0) {
		printf("ERROR: read d88 image '%s'\n", img_file);
		goto ERROR_EXIT;
	}

	if ((disk = get_d88_disk(&image, image_index)) == NULL) {
		printf("ERROR: illegal image index: '%s':%d\n", img_file, image_index);
		goto ERROR_EXIT;
	}

	for (n = 0; n < 2; n++) {
		int i = ids[n];
		int j;
		int res;

		for (j = 0; j < 2; j++) {
			const struct d88_sector_t *sector;

			if ((sector = get_d88_sector(disk,
										 sec_info[i][j].cylinder,
										 sec_info[i][j].head,
										 sec_info[i][j].record)) == NULL ||
				sector->N != 1) {
				printf("ERROR: read d88 image '%s':%d %02X/%02X/%02X\n",
					   img_file,
					   image_index,
					   sec_info[i][j].cylinder,
					   sec_info[i][j].head,
					   sec_info[i][j].record);
				goto ERROR_EXIT;
			}

			memcpy(&(map_data[j * 4][0][0]), sector->data, 256);
		}

		if ((res = modify_map_data(map_data, i)) < 0) {
			printf("ERROR: can't modify map\n");
			goto ERROR_EXIT;
		} else if (res == 0) {
			continue;
		}

		for (j = 0; j < 2; j++) {
			struct d88_sector_t *sector;

			if ((sector = get_d88_sector(disk,
										 sec_info[i][j].cylinder,
										 sec_info[i][j].head,
										 sec_info[i][j].record)) == NULL ||
				sector->N != 1) {
				printf("ERROR: read d88 image '%s':%d %02X/%02X/%02X\n",
					   img_file,
					   image_index,
					   sec_info[i][j].cylinder,
					   sec_info[i][j].head,
					   sec_info[i][j].record);
				goto ERROR_EXIT;
			}

			if (memcmp(sector->data, &(map_data[j * 4][0][0]), 256) != 0) {
				memcpy(sector->data, &(map_data[j * 4][0][0]), 256);
				modified = 1;
			}
		}
	}

	if (modified) {
		if (opt_overwrite == 0 && check_out_file(out_file) < 0) {
			goto ERROR_EXIT;
		}

		if (write_d88_image(&image, out_file) < 0) {
			printf("ERROR: write d88 image '%s'\n", out_file);
			goto ERROR_EXIT;
		}
	} else {
		printf("already modified\n");
	}

	result = modified;

ERROR_EXIT:

	free_d88_image(&image, 0);

	return result;
}
#endif	/* U4_PORT_PC88 */

#if defined(U4_PORT_PC98) || defined(U4_PORT_X68K)
static int modify_map_data_dosfd(const char *img_file, const char *out_file, int offset, int opt_overwrite)
{
	int result = -1;
	const char *data_file = "DUNGEON.BIN";
	struct dosfd_image_t image;
	unsigned char map_data[DNG_ID_MAX][8][8][8];		/* map_data[dungeon][level][y][x] */
	const int ids[2] = { DNG_ID_WRONG, DNG_ID_COVETOUS };
	int n;
	int modified = 0;

	if (read_dosfd_image(&image, img_file, offset) < 0) {
		printf("ERROR: read disk image '%s'\n", img_file);
		goto ERROR_EXIT;
	}

	if (read_dosfd_image_file(&image, data_file, (unsigned char *)map_data, sizeof(map_data), 1) < 0) {
		printf("ERROR: read disk image '%s/%s'\n", img_file, data_file);
		goto ERROR_EXIT;
	}

	for (n = 0; n < 2; n++) {
		int i = ids[n];
		int res;

		if ((res = modify_map_data(map_data[i], i)) < 0) {
			printf("ERROR: can't modify map\n");
			goto ERROR_EXIT;
		} else if (res > 0) {
			modified = 1;
		}
	}

	if (modified) {
		if (opt_overwrite == 0 && check_out_file(out_file) < 0) {
			goto ERROR_EXIT;
		}

		if (write_dosfd_image_file(&image, data_file, (unsigned char *)map_data, sizeof(map_data), 1) < 0) {
			printf("ERROR: write disk image '%s/%s'\n", out_file, data_file);
			goto ERROR_EXIT;
		}

		if (write_dosfd_image(&image, out_file) < 0) {
			printf("ERROR: write disk image '%s'\n", out_file);
			goto ERROR_EXIT;
		}
	} else {
		printf("already modified\n");
	}

	result = modified;

ERROR_EXIT:
	free_dosfd_image(&image, 0);

	return result;
}
#endif	/* U4_PORT_PC98 || U4_PORT_X68K */

static int modify_map_data(unsigned char map_data[8][8][8], int dng_id)
{
	int result = -1;

	if (dng_id == DNG_ID_WRONG) {
		printf("Wrong L4:\n");
		dump_map_data(map_data[3]);
		if (map_data[3][3][7] == 0x10 && map_data[3][4][7] == 0x40) {
			map_data[3][3][7] = 0x40;	/* 上りハシゴ → 宝箱       */
			map_data[3][4][7] = 0x10;	/* 宝箱       → 上りハシゴ */

			printf("modified:\n");
			dump_map_data(map_data[3]);

			result = 1;
		} else if (map_data[3][3][7] == 0x40 && map_data[3][4][7] == 0x10) {
			result = 0;
		} else {
			printf("ERROR: wrong data\n");
		}
	} else if (dng_id == DNG_ID_COVETOUS) {
		printf("Covetous L7:\n");
		dump_map_data(map_data[6]);
		if (map_data[6][3][3] == 0x00) {
			map_data[6][3][3] = 0x10;	/* 通路       → 上りハシゴ */

			printf("modified:\n");
			dump_map_data(map_data[6]);

			result = 1;
		} else if (map_data[6][3][3] == 0x10) {
			result = 0;
		} else {
			printf("ERROR: wrong data\n");
		}
	} else {
		result = 0;
	}

	return result;
}

static void dump_map_data(unsigned char map_data[8][8])
{
	int x;
	int y;
	for (y = 0; y < 8; y++) {
		for (x = 0; x < 8; x++) {
			if (x == 0) {
				printf("\t");
			} else {
				printf(" ");
			}
			printf("%02X", (int)map_data[y][x]);
		}
		printf("\n");
	}
}

static char *create_new_name(const char *name)
{
	int baselen;
	const char *ext;
	int len = strlen(name) + 5;
	char *new_name = (char *)malloc(len);
	if (new_name == NULL) {
		printf("ERROR: memory exhausted!\n");
		return NULL;
	}

	if ((ext = strrchr(name, '.')) == NULL) {
		baselen = strlen(name);
		ext = "";
	} else {
		baselen = ext - name;
	}

	strncpy(new_name, name, baselen);
	strcpy(new_name + baselen, ".NEW");
	strcpy(new_name + baselen + 4, ext);

	return new_name;
}

static int check_out_file(const char *filename)
{
	struct stat st;
	if (stat(filename, &st) == 0) {
		printf("already exists '%s'\n", filename);
		if (confirm("overwrite ?") <= 0) {
			printf("aborted!\n");
			return -1;
		}
	}

	return 0;
}

static int confirm(const char *msg)
{
	printf("%s (Y or N): ", msg);
	fflush(stdout);

	while (1) {
		int c = getchar();
		if (c == 'y' || c == 'Y') {
			return 1;
		}
		if (c == 'n' || c == 'N') {
			return 0;
		}
	}

	return -1;
}

/* EOF */
